#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <setjmp.h>

typedef struct Displaylist {
  struct Displaylist * next;
  char *               dpyname;
  Display *            dpy;
  Widget               top;
  Widget               w;
} Displaylist;
  
typedef struct Global {
  Displaylist *   displays;  /* display list header */
  XtAppContext    app;       /* application context */
  FILE *          pipe;      /* named pipe from which to read */
  XtInputId       inputid;   /* input handler id */
  char *          argv0;     /* program name */
} Global;

Global global;

jmp_buf xio_recover;

/* functions */
void Close ();
void Input ();
void nullTimerProc ();
char * ParseDisplay ();
char * ParseMessage ();

static XtActionsRec actions [] = {
  {"Close", (XtActionProc) Close}, 
};

/*ARGSUSED*/
void Close (w, ev, params, num_params)
     Widget w;
     XEvent * ev;
     String params[];
     Cardinal * num_params;
{
  destroyDisplay (XtDisplay(w));
}


destroyDisplay (dpy)
     Display * dpy;
{
  Displaylist * dl;
  Displaylist ** prev;

  dl = global.displays;
  prev = &global.displays;

  while (dl) {
    if (dl->dpy == dpy) {
      /* this is the one to wipe out */
      XtDestroyWidget (dl->top);
      XtCloseDisplay (dpy);
      if (dl->dpyname) XtFree (dl->dpyname);
      /* unlink this entry */
      *prev = dl->next;
      XtFree (dl);
      break;      
    }
    prev = &dl->next;
    dl = dl->next;
  }
}

xio_error_handler(dpy)
     Display * dpy;
{
  destroyDisplay (dpy);
  longjmp (xio_recover, 1);
}

main (argc, argv)
  int    argc;
  char  *argv[];
{
  char *        pipename;

  /*
   * check parameters
   */
  switch (argc) {
  case 1: pipename = "/tmp/notice"; break;
  case 2: pipename = argv[1]; break;
  default:
    fprintf (stderr,"Usage: %s [pipename]\n", argv[0]);
    exit (1);
    break;
  }
  global.argv0 = argv[0];

  /*
   * Initialize the intrinsics
   */
  XtToolkitInitialize ();
  global.app = XtCreateApplicationContext ();
  XtAppAddActions(global.app, actions, XtNumber(actions));
  
  /*
   * Open the named pipe for reading
   */
  global.pipe = fopen (pipename, "r");
  if (!global.pipe) {
    fprintf (stderr, "%s: %s not found\n", 
	     argv[0], pipename);
    exit(1);
  }
  /*
   * register the input handler
   */
  global.inputid = XtAppAddInput
                    (global.app,
		     fileno(global.pipe),
		     XtInputReadMask,
		     Input, &global);

  /*
   * error recovery, and main loop
   */
  (void) setjmp (xio_recover);
  XSetIOErrorHandler (xio_error_handler);
  XtAppMainLoop(global.app);
}

void
Input (global, fd, inputid)
     Global * global;
     int * fd;
     XtInputId * inputid;
{
  char buf [BUFSIZ];
  char * dpyname;
  Displaylist * dl = NULL;
  Displaylist * dlp;
  Arg labelarg;

  if (*fd == fileno(global->pipe)) {
    /* read, handle possible error or EOF condition */
    if (!fgets (buf, BUFSIZ, global->pipe)) 
      XtRemoveInput(*inputid);
    buf[strlen(buf)-1] = '\0'; /* chop line */
    dpyname = ParseDisplay (buf);
    XtSetArg (labelarg, XtNlabel,
	      (XtArgVal)ParseMessage(buf));

    for (dlp = global->displays;
	 dlp;
	 dlp = dlp->next ) {
      if (0 == strcmp(dpyname,
		      dlp->dpyname)) {
	dl = dlp;
      }
    }

    if (!dl) {
      Display *dpy;
      int argc = 1;
      char * argv = global->argv0;
      Atom wm_delete_window;

      dpy= XtOpenDisplay
	     (global->app,
	      dpyname,
	      NULL, "Multi",
	      NULL, 0,
	      &argc, &argv);

      if (!dpy) return;

      dl = (Displaylist * ) XtCalloc (1,sizeof (Displaylist));
      if (!dl) abort();

      dl->dpy = dpy;
      dl->dpyname = dpyname;
      dl->top = XtAppCreateShell 
                    (NULL, "Multi",
		     applicationShellWidgetClass,
		     dpy, NULL, 0);

      dl->w = XtCreateManagedWidget 
                    ("label", labelWidgetClass,
	             dl->top, 
		     &labelarg, 1L);
  
      XtRealizeWidget (dl->top);

      wm_delete_window = XInternAtom
	                   (XtDisplay(dl->top), 
			    "WM_DELETE_WINDOW", False);
      (void) XSetWMProtocols 
	        (XtDisplay(dl->top), XtWindow(dl->top),
		 &wm_delete_window, 1);

      XtOverrideTranslations
         (dl->top,
	  XtParseTranslationTable
              ("<Message>WM_PROTOCOLS: Close()"));

      dl->next = global->displays;
      global->displays = dl;
    } else {
      XtSetValues (dl->w, &labelarg, 1L);
    }
    (void) XtAppAddTimeOut(global->app, 1L, nullTimerProc, global);
  }
}

void
nullTimerProc ()
{
}


char *
ParseDisplay (buf)
     char * buf;
{
  char * res;
  char * l = (char * ) strchr(buf,'#');
  int len;

  if (l) {
    len = l - buf;
    res = XtCalloc(1,len+1);
    if (!res) abort();
    if(len) strncpy (res,buf,len);
  } else {
    res = XtCalloc(1,len+1);
  }
  return res;
}		 
  
char *
ParseMessage (buf)
     char * buf;
{
  char * l = (char *) strchr(buf,'#');

  if (l) l++;
  else l = buf;
  return l;

}
