/*
 *  wdiag  --  widget hierarchy diagnostics.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
#include <X11/CompositeP.h>
#include <ctype.h>

char * nameof (widget)
     Widget widget;
{
  char * valstr;
  static char fulname[1024];
  Widget ancestry[256];
  Widget w;
  int nancs;
  int i;
  char ** str_type;

  nancs = 0;
  /* trace widget's ancestry */
  for (w=widget; w; w=w->core.parent) {
    ancestry[nancs++] = w;
  }
  /* run backwards through ancestry elucidating name string*/
  fulname[0]='\0';
  for (i = nancs; i; ) {
    --i;
    w = ancestry[i];
    strcat (fulname,w->core.name);
    if (i) {
      strcat (fulname, ".");
    }
  }
  return fulname;
}
  
char * classof (widget)
     Widget widget;
{
  char * valstr;
  static char fulclas[1024];
  Widget ancestry[256];
  Widget w;
  int nancs;
  int i;
  char ** str_type;

  nancs = 0;
  /* trace widget's ancestry */
  for (w=widget; w; w=w->core.parent) {
    ancestry[nancs++] = w;
  }
  /* run backwards through ancestry elucidating strings */
  fulclas[0]='\0';
  for (i = nancs; i; ) {
    --i;
    w = ancestry[i];
    strcat (fulclas,w->core.widget_class->core_class.class_name);
    if (i) {
      strcat (fulclas, ".");
    }
  }
  return fulclas;
}
  
Widget rootof (widget)
     Widget widget;
{
  Widget w;
  Widget result;

  /* trace widget's ancestry */
  for (w=widget; w; w=w->core.parent) {
    result = w;
  }
  return result;
}
  
void treeof (widget)
     Widget widget;
{
  char * valstr;
  Widget * wp;
  int wc;
  int i;

  printf ("%-40s %s\n",nameof(widget),classof(widget));
  if (XtIsComposite(widget)) {
    CompositeRec * compw = (CompositeRec *) widget;
    wp = compw->composite.children;
    wc = compw->composite.num_children;
    for (i = 0; i < wc ; i++) {
      treeof (wp[i]);
    }
  }
}
  
char * valueof (widget, value)
     Widget widget;
     char * value;
{
  char name[1024];
  char clas[1024];
  char val [1024];
  char str [1024];
  XrmValue rmvalue;
  char * str_type;
  Display * dpy = XtDisplay(widget);
  XrmDatabase database = dpy->db;
  Boolean found;

  strcpy (name, nameof(widget));
  strcat (name, ".");
  strcat (name, value);

  strcpy (val, value);
  if (islower(val[0])) val[0]=toupper(val[0]);
  strcpy (clas, classof(widget));
  strcat (clas, ".");
  strcat (clas, val);

  found = XrmGetResource (database, name, clas,
			  &str_type, &rmvalue);

  val[0] = '\0';

  if (found) {
    
    if (0==strcmp ("String", str_type)) {
      sprintf (val, "(%s)", str_type);
      strncpy (str, rmvalue.addr, rmvalue.size);
      str[rmvalue.size] = '\0';
      strcat (val," ");
      strcat (val,str);
    }
    else if (   0==strcmp(str_type, "Dimension") 
	     || 0==strcmp(str_type, "Int")
	     || 0==strcmp(str_type, "Position")) {
      sprintf (val, "(%s) %d", str_type, *rmvalue.addr);
    }
    else {
      sprintf (val, "(%s) (length = %d)", str_type, rmvalue.size);
    }      
  }
  return val;
}








